@@ -18,6 +18,7 @@ try:
 except ImportError:
     from chameleon import ast24 as ast
 
+import sys
 import logging
 import weakref
 
@@ -36,6 +37,12 @@ except TypeError:
 __docformat__ = 'restructuredtext en'
 
 
+def annotated(value):
+    node = load("annotation")
+    node_annotations[node] = value
+    return node
+
+
 def parse(source, mode='eval'):
     return compile(source, '', mode, ast.PyCF_ONLY_AST)
 
@@ -85,6 +92,10 @@ def swap(root, replacement, name):
             node_annotations.setdefault(node, replacement)
 
 
+def marker(name):
+    return ast.Str(s="__%s" % name)
+
+
 class Node(object):
     """AST baseclass that gives us a convenient initialization
     method. We explicitly declare and use the ``_fields`` attribute."""
@@ -152,9 +163,9 @@ class ASTCodeGenerator(object):
 
     def __init__(self, tree):
         self.lines_info = []
-        self.line_info = None
+        self.line_info = []
         self.lines = []
-        self.line = None
+        self.line = ""
         self.last = None
         self.indent = 0
         self.blame_stack = []
@@ -200,6 +211,10 @@ class ASTCodeGenerator(object):
                 self.line_info.append((len(self.line), self.last))
         self.line += s
 
+    def flush(self):
+        if self.line:
+            self._new_line()
+
     def visit(self, node):
         if node is None:
             return None
@@ -415,7 +430,11 @@ class ASTCodeGenerator(object):
         self._new_line()
         self._write('raise')
         if not getattr(node, "type", None):
-            return
+            exc = getattr(node, "exc", None)
+            if exc is None:
+                return
+            self._write(' ')
+            return self.visit(exc)
         self._write(' ')
         self.visit(node.type)
         if not node.inst:
@@ -427,6 +446,34 @@ class ASTCodeGenerator(object):
         self._write(', ')
         self.visit(node.tback)
 
+    # Try(stmt* body, excepthandler* handlers, stmt* orelse, stmt* finalbody)
+    def visit_Try(self, node):
+        self._new_line()
+        self._write('try:')
+        self._change_indent(1)
+        for statement in node.body:
+            self.visit(statement)
+        self._change_indent(-1)
+        if getattr(node, 'handlers', None):
+            for handler in node.handlers:
+                self.visit(handler)
+        self._new_line()
+
+        if getattr(node, 'orelse', None):
+            self._write('else:')
+            self._change_indent(1)
+            for statement in node.orelse:
+                self.visit(statement)
+            self._change_indent(-1)
+
+        if getattr(node, 'finalbody', None):
+            self._new_line()
+            self._write('finally:')
+            self._change_indent(1)
+            for statement in node.finalbody:
+                self.visit(statement)
+            self._change_indent(-1)
+
     # TryExcept(stmt* body, excepthandler* handlers, stmt* orelse)
     def visit_TryExcept(self, node):
         self._new_line()
@@ -454,7 +501,11 @@ class ASTCodeGenerator(object):
             self._write(' ')
             self.visit(node.type)
         if getattr(node, 'name', None):
-            self._write(', ')
+            if sys.version_info[0] == 2:
+                assert getattr(node, 'type', None)
+                self._write(', ')
+            else:
+                self._write(' as ')
             self.visit(node.name)
         self._write(':')
         self._change_indent(1)
@@ -640,6 +691,16 @@ class ASTCodeGenerator(object):
             self._write(', ')
         self._write('}')
 
+    def visit_Set(self, node):
+        self._write('{')
+        elts = list(node.elts)
+        last = elts.pop()
+        for elt in elts:
+            self.visit(elt)
+            self._write(', ')
+        self.visit(last)
+        self._write('}')
+
     # ListComp(expr elt, comprehension* generators)
     def visit_ListComp(self, node):
         self._write('[')
@@ -813,9 +874,10 @@ class AnnotationAwareVisitor(ast.NodeVisitor):
         super(AnnotationAwareVisitor, self).visit(node)
 
     def apply_transform(self, node):
-        result = self.transform(node)
-        if result is not node:
-            node_annotations[node] = result
+        if node not in node_annotations:
+            result = self.transform(node)
+            if result is not None and result is not node:
+                node_annotations[node] = result
 
 
 class NameLookupRewriteVisitor(AnnotationAwareVisitor):
@@ -844,13 +906,11 @@ class NameLookupRewriteVisitor(AnnotationAwareVisitor):
         finally:
             self.scopes.pop()
 
+
 class ItemLookupOnAttributeErrorVisitor(AnnotationAwareVisitor):
     def __init__(self, transform):
         self.transform = transform
 
     def visit_Attribute(self, node):
         self.generic_visit(node)
-
-        # Only apply transform for name values
-        if isinstance(node.value, ast.Name) and isinstance(node.ctx, ast.Load):
-            self.apply_transform(node)
+        self.apply_transform(node)
