@@ -1,3 +1,5 @@
+import re
+
 try:
     import ast
 except ImportError:
@@ -21,6 +23,7 @@ from ..namespaces import META_NS as META
 
 from ..astutil import Static
 from ..astutil import parse
+from ..astutil import marker
 
 from .. import tal
 from .. import metal
@@ -31,7 +34,7 @@ from ..exc import LanguageError
 from ..exc import ParseError
 from ..exc import CompilationError
 
-from ..utils import unescape
+from ..utils import decode_htmlentities
 
 try:
     str = unicode
@@ -55,7 +58,9 @@ def wrap(node, *wrappers):
 def validate_attributes(attributes, namespace, whitelist):
     for ns, name in attributes:
         if ns == namespace and name not in whitelist:
-            raise CompilationError("Bad attribute for namespace '%s'" % ns, name)
+            raise CompilationError(
+                "Bad attribute for namespace '%s'" % ns, name
+                )
 
 
 class MacroProgram(ElementProgram):
@@ -82,6 +87,32 @@ class MacroProgram(ElementProgram):
     # Macro name (always trivial for a macro program)
     name = None
 
+    # This default marker value has the semantics that if an
+    # expression evaluates to that value, the expression default value
+    # is returned. For an attribute, if there is no default, this
+    # means that the attribute is dropped.
+    default_marker = None
+
+    # Escape mode (true value means XML-escape)
+    escape = True
+
+    # Attributes which should have boolean behavior (on true, the
+    # value takes the attribute name, on false, the attribute is
+    # dropped)
+    boolean_attributes = set()
+
+    # If provided, this should be a set of attributes for implicit
+    # translation. Any attribute whose name is included in the set
+    # will be translated even without explicit markup. Note that all
+    # values should be lowercase strings.
+    implicit_i18n_attributes = set()
+
+    # If set, text will be translated even without explicit markup.
+    implicit_i18n_translate = False
+
+    # If set, additional attribute whitespace will be stripped.
+    trim_attribute_space = False
+
     def __init__(self, *args, **kwargs):
         # Internal array for switch statements
         self._switches = []
@@ -95,16 +126,27 @@ class MacroProgram(ElementProgram):
         # Internal dictionary of macro definitions
         self._macros = {}
 
-        # Set escape mode (true value means XML-escape)
-        self._escape = kwargs.pop('escape', True)
+        # Apply default values from **kwargs to self
+        self._pop_defaults(
+            kwargs,
+            'boolean_attributes',
+            'default_marker',
+            'escape',
+            'implicit_i18n_translate',
+            'implicit_i18n_attributes',
+            'trim_attribute_space',
+            )
 
         super(MacroProgram, self).__init__(*args, **kwargs)
 
     @property
     def macros(self):
+        macros = list(self._macros.items())
+        macros.append((None, nodes.Sequence(self.body)))
+
         return tuple(
-            nodes.Macro(name, [node])
-            for name, node in self._macros.items()
+            nodes.Macro(name, [nodes.Context(node)])
+            for name, node in macros
             )
 
     def visit_default(self, node):
@@ -113,6 +155,10 @@ class MacroProgram(ElementProgram):
     def visit_element(self, start, end, children):
         ns = start['ns_attrs']
 
+        for (prefix, attr), encoded in tuple(ns.items()):
+            if prefix == TAL:
+                ns[prefix, attr] = decode_htmlentities(encoded)
+
         # Validate namespace attributes
         validate_attributes(ns, TAL, tal.WHITELIST)
         validate_attributes(ns, METAL, metal.WHITELIST)
@@ -134,7 +180,7 @@ class MacroProgram(ElementProgram):
         except KeyError:
             switch = None
         else:
-            switch = nodes.Expression(clause)
+            switch = nodes.Value(clause)
 
         self._switches.append(switch)
 
@@ -149,11 +195,11 @@ class MacroProgram(ElementProgram):
 
             if use_macro:
                 inner = nodes.UseExternalMacro(
-                    nodes.Expression(use_macro), slots, False
+                    nodes.Value(use_macro), slots, False
                     )
             else:
                 inner = nodes.UseExternalMacro(
-                    nodes.Expression(extend_macro), slots, True
+                    nodes.Value(extend_macro), slots, True
                     )
         # -or- include tag
         else:
@@ -166,11 +212,8 @@ class MacroProgram(ElementProgram):
                 pass
             else:
                 key, value = tal.parse_substitution(clause)
-                value = unescape(value)
-                expression = nodes.Expression(value)
-                msgid = ns.get((I18N, 'translate'))
-                content = self._make_content_node(
-                    expression, msgid, key, content)
+                xlate = True if ns.get((I18N, 'translate')) == '' else False
+                content = self._make_content_node(value, content, key, xlate)
 
                 if end is None:
                     # Make sure start-tag has opening suffix.
@@ -190,8 +233,7 @@ class MacroProgram(ElementProgram):
             except KeyError:
                 pass
             else:
-                dynamic = (ns.get((TAL, 'content')) or \
-                           ns.get((TAL, 'replace')))
+                dynamic = ns.get((TAL, 'content')) or ns.get((TAL, 'replace'))
 
                 if not dynamic:
                     content = nodes.Translate(clause, content)
@@ -214,7 +256,8 @@ class MacroProgram(ElementProgram):
 
             # Prepare attributes from TAL language
             prepared = tal.prepare_attributes(
-                start['attrs'], TAL_ATTRIBUTES, ns, self.DROP_NS
+                start['attrs'], TAL_ATTRIBUTES,
+                I18N_ATTRIBUTES, ns, self.DROP_NS
                 )
 
             # Create attribute nodes
@@ -226,16 +269,16 @@ class MacroProgram(ElementProgram):
             # Start- and end nodes
             start_tag = nodes.Start(
                 start['name'],
-                start['prefix'],
-                start['suffix'],
+                self._maybe_trim(start['prefix']),
+                self._maybe_trim(start['suffix']),
                 ATTRIBUTES
                 )
 
             end_tag = nodes.End(
                 end['name'],
                 end['space'],
-                end['prefix'],
-                end['suffix'],
+                self._maybe_trim(end['prefix']),
+                self._maybe_trim(end['suffix']),
                 ) if end is not None else None
 
             # tal:omit-tag
@@ -249,7 +292,7 @@ class MacroProgram(ElementProgram):
                 if clause == "":
                     omit = True
                 else:
-                    expression = nodes.Negate(nodes.Expression(clause))
+                    expression = nodes.Negate(nodes.Value(clause))
                     omit = expression
 
                     # Wrap start- and end-tags in condition
@@ -269,7 +312,7 @@ class MacroProgram(ElementProgram):
 
                 # Assign static attributes dictionary to "attrs" value
                 inner = nodes.Define(
-                    [nodes.Assignment(["attrs"], STATIC_ATTRIBUTES, True)],
+                    [nodes.Alias(["attrs"], STATIC_ATTRIBUTES)],
                     inner,
                     )
 
@@ -283,10 +326,8 @@ class MacroProgram(ElementProgram):
                 pass
             else:
                 key, value = tal.parse_substitution(clause)
-                value = unescape(value)
-                expression = nodes.Expression(value)
-                msgid = ns.get((I18N, 'translate'))
-                inner = self._make_content_node(expression, msgid, key, inner)
+                xlate = True if ns.get((I18N, 'translate')) == '' else False
+                inner = self._make_content_node(value, inner, key, xlate)
 
         # metal:define-slot
         try:
@@ -309,7 +350,7 @@ class MacroProgram(ElementProgram):
             DEFINE = partial(
                 nodes.Define,
                 [nodes.Assignment(
-                    names, nodes.Expression(expr), context == "local")
+                    names, nodes.Value(expr), context == "local")
                  for (context, names, expr) in defines],
                 )
 
@@ -319,7 +360,7 @@ class MacroProgram(ElementProgram):
         except KeyError:
             CASE = skip
         else:
-            value = nodes.Expression(clause)
+            value = nodes.Value(clause)
             for switch in reversed(self._switches):
                 if switch is not None:
                     break
@@ -346,7 +387,7 @@ class MacroProgram(ElementProgram):
             assert len(defines) == 1
             context, names, expr = defines[0]
 
-            expression = nodes.Expression(expr)
+            expression = nodes.Value(expr)
 
             REPEAT = partial(
                 nodes.Repeat,
@@ -362,7 +403,7 @@ class MacroProgram(ElementProgram):
         except KeyError:
             CONDITION = skip
         else:
-            expression = nodes.Expression(clause)
+            expression = nodes.Value(clause)
             CONDITION = partial(nodes.Condition, expression)
 
         # tal:switch
@@ -371,6 +412,22 @@ class MacroProgram(ElementProgram):
         else:
             SWITCH = partial(nodes.Cache, [switch])
 
+        # i18n:domain
+        try:
+            clause = ns[I18N, 'domain']
+        except KeyError:
+            DOMAIN = skip
+        else:
+            DOMAIN = partial(nodes.Domain, clause)
+
+        # i18n:name
+        try:
+            clause = ns[I18N, 'name']
+        except KeyError:
+            NAME = skip
+        else:
+            NAME = partial(nodes.Name, clause)
+
         # The "slot" node next is the first node level that can serve
         # as a macro slot
         slot = wrap(
@@ -381,6 +438,7 @@ class MacroProgram(ElementProgram):
             CONDITION,
             REPEAT,
             SWITCH,
+            DOMAIN,
             )
 
         # metal:fill-slot
@@ -389,7 +447,7 @@ class MacroProgram(ElementProgram):
         except KeyError:
             pass
         else:
-            index =-(1 + int(bool(use_macro or extend_macro)))
+            index = -(1 + int(bool(use_macro or extend_macro)))
 
             try:
                 slots = self._use_macro[index]
@@ -411,21 +469,10 @@ class MacroProgram(ElementProgram):
             self._macros[clause] = slot
             slot = nodes.UseInternalMacro(clause)
 
-        # i18n:domain
-        try:
-            clause = ns[I18N, 'domain']
-        except KeyError:
-            DOMAIN = skip
-        else:
-            DOMAIN = partial(nodes.Domain, clause)
-
-        # i18n:name
-        try:
-            clause = ns[I18N, 'name']
-        except KeyError:
-            NAME = skip
-        else:
-            NAME = partial(nodes.Name, clause)
+        slot = wrap(
+            slot,
+            NAME
+            )
 
         # tal:on-error
         try:
@@ -433,9 +480,19 @@ class MacroProgram(ElementProgram):
         except KeyError:
             ON_ERROR = skip
         else:
-            expression = nodes.Expression(clause)
-            fallback = nodes.Content(expression, None, False)
-            ON_ERROR = partial(nodes.OnError, fallback)
+            key, value = tal.parse_substitution(clause)
+            translate = True if ns.get((I18N, 'translate')) == '' else False
+
+            fallback = self._make_content_node(value, None, key, translate)
+
+            if omit is False and start['namespace'] not in self.DROP_NS:
+                fallback = nodes.Element(
+                start_tag,
+                end_tag,
+                fallback,
+                )
+
+            ON_ERROR = partial(nodes.OnError, fallback, 'error')
 
         clause = ns.get((META, 'interpolation'))
         if clause in ('false', 'off'):
@@ -461,34 +518,80 @@ class MacroProgram(ElementProgram):
 
         return wrap(
             slot,
-            DOMAIN,
-            NAME,
             ON_ERROR
             )
 
     def visit_start_tag(self, start):
         return self.visit_element(start, None, [])
 
+    def visit_cdata(self, node):
+        if not self._interpolation[-1] or not '${' in node:
+            return nodes.Text(node)
+
+        expr = nodes.Substitution(node, ())
+        return nodes.Interpolation(expr, True, False)
+
     def visit_comment(self, node):
         if node.startswith('<!--!'):
             return
 
-        if self._interpolation[-1]:
-            return nodes.Sequence(
-                [nodes.Text(node[:4]),
-                 nodes.Interpolation(node[4:-3], self._escape, True),
-                 nodes.Text(node[-3:])
-                 ])
+        if node.startswith('<!--?'):
+            return nodes.Text('<!--' + node.lstrip('<!-?'))
 
-        return nodes.Text(node)
+        if not self._interpolation[-1] or not '${' in node:
+            return nodes.Text(node)
+
+        char_escape = ('&', '<', '>') if self.escape else ()
+        expression = nodes.Substitution(node[4:-3], char_escape)
+
+        return nodes.Sequence(
+            [nodes.Text(node[:4]),
+             nodes.Interpolation(expression, True, False),
+             nodes.Text(node[-3:])
+             ])
 
     def visit_text(self, node):
         self._last = node
 
-        if self._interpolation_enabled:
-            return nodes.Interpolation(node, self._escape, True)
+        translation = self.implicit_i18n_translate
 
-        return nodes.Text(node)
+        if self._interpolation[-1] and '${' in node:
+            char_escape = ('&', '<', '>') if self.escape else ()
+            expression = nodes.Substitution(node, char_escape)
+            return nodes.Interpolation(expression, True, translation)
+
+        if not translation:
+            return nodes.Text(node)
+
+        seq = []
+        while node:
+            m = re.search('\s+', node)
+            if m is not None:
+                s = m.start()
+                if s:
+                    t = node[:s]
+                    seq.append(nodes.Translate(t, nodes.Text(t)))
+
+                e = m.end()
+                whitespace = nodes.Text(node[s:e])
+                seq.append(whitespace)
+
+                if e < len(node):
+                    node = node[e:]
+                    continue
+
+            else:
+                seq.append(nodes.Translate(node, nodes.Text(node)))
+
+            break
+
+        return nodes.Sequence(seq)
+
+    def _pop_defaults(self, kwargs, *attributes):
+        for attribute in attributes:
+            default = getattr(self, attribute)
+            value = kwargs.pop(attribute, default)
+            setattr(self, attribute, value)
 
     def _check_attributes(self, namespace, ns):
         if namespace in self.DROP_NS and ns.get((TAL, 'attributes')):
@@ -516,64 +619,88 @@ class MacroProgram(ElementProgram):
                 tal_content
                 )
 
-    def _make_content_node(self, expression, msgid, key, default):
-        content = nodes.Content(expression, msgid, key == "text")
+    def _make_content_node(self, expression, default, key, translate):
+        value = nodes.Value(expression)
+        char_escape = ('&', '<', '>') if key == 'text' else ()
+        content = nodes.Content(value, char_escape, translate)
 
-        content = nodes.Condition(
-            nodes.Identity(expression, nodes.Marker("default")),
-            default,
-            content,
-            )
+        if default is not None:
+            content = nodes.Condition(
+                nodes.Identity(value, marker("default")),
+                default,
+                content,
+                )
 
-        # Cache expression to avoid duplicate evaluation
-        content = nodes.Cache([expression], content)
+            # Cache expression to avoid duplicate evaluation
+            content = nodes.Cache([value], content)
 
-        # Define local marker "default"
-        content = nodes.Define(
-            [nodes.Assignment(["default"], nodes.Marker("default"), True)],
-            content
-            )
+            # Define local marker "default"
+            content = nodes.Define(
+                [nodes.Alias(["default"], marker("default"))],
+                content
+                )
 
         return content
 
     def _create_attributes_nodes(self, prepared, I18N_ATTRIBUTES):
         attributes = []
-        for name, text, quote, space, eq, expr in prepared:
-            # If (by heuristic) ``text`` contains one or
-            # more interpolation expressions, make the attribute
-            # dynamic
-            if expr is None and text is not None and '${' in text:
-                value = nodes.Interpolation(text, True, True)
 
-            # If this expression is non-trivial, the attribute is
-            # dynamic (computed)
+        for name, text, quote, space, eq, expr in prepared:
+            implicit_i18n = name.lower() in self.implicit_i18n_attributes
+
+            char_escape = ('&', '<', '>', quote)
+
+            # Use a provided default text as the default marker
+            # (aliased to the name ``default``), otherwise use the
+            # program's default marker value.
+            if text is not None:
+                default_marker = ast.Str(s=text)
+            else:
+                default_marker = self.default_marker
+
+            msgid = I18N_ATTRIBUTES.get(name, missing)
+
+            # If (by heuristic) ``text`` contains one or more
+            # interpolation expressions, apply interpolation
+            # substitution to the text
+            if expr is None and text is not None and '${' in text:
+                expr = nodes.Substitution(text, char_escape, None)
+                translation = implicit_i18n and msgid is missing
+                value = nodes.Interpolation(expr, True, translation)
+                default_marker = self.default_marker
+
+            # If the expression is non-trivial, the attribute is
+            # dynamic (computed).
             elif expr is not None:
-                value = nodes.Expression(expr)
+                if name in self.boolean_attributes:
+                    value = nodes.Boolean(expr, name)
+                else:
+                    if text is not None:
+                        default = default_marker
+                    else:
+                        default = None
+
+                    value = nodes.Substitution(expr, char_escape, default)
 
             # Otherwise, it's a static attribute.
             else:
                 value = ast.Str(s=text)
+                if msgid is missing and implicit_i18n:
+                    msgid = text
 
             # If translation is required, wrap in a translation
             # clause
-            msgid = I18N_ATTRIBUTES.get(name, missing)
             if msgid is not missing:
                 value = nodes.Translate(msgid, value)
 
-            # Escape non-static values
-            escape = not isinstance(value, ast.Str)
+            space = self._maybe_trim(space) or " "
+            attribute = nodes.Attribute(name, value, quote, eq, space)
 
-            attribute = nodes.Attribute(name, value, quote, eq, space, escape)
-
-            # If value is non-static, wrap attribute in a definition
-            # clause for the "default" value
-            if not isinstance(value, ast.Str):
-                default = ast.Str(s=text) if text is not None \
-                          else ast.Name(id="None", ctx=ast.Load())
-                attribute = nodes.Define(
-                    [nodes.Assignment(["default"], default, True)],
-                    attribute,
-                    )
+            # Always define a ``default`` alias
+            attribute = nodes.Define(
+                [nodes.Alias(["default"], default_marker)],
+                attribute,
+                )
 
             attributes.append(attribute)
 
@@ -586,3 +713,9 @@ class MacroProgram(ElementProgram):
             static_attrs[name] = text if text is not None else expr
 
         return Static(parse(repr(static_attrs)).body)
+
+    def _maybe_trim(self, string):
+        if self.trim_attribute_space:
+            return string.strip()
+
+        return string
